#!/usr/bin/perl -w
#############################################################################
### BlockIn: An OpenGL clone of BlockOut
### Copyright 2004 Joseph Jezak
#############################################################################

#############################################################################
### This file is part of Blockin.
###
### BlockIn is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.
###
### BlockIn is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.
###
### You should have received a copy of the GNU General Public License
### along with BlockIn; if not, write to the Free Software
### Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

my $version = "0.06";
print "BlockIn v$version\n";

### Use strict, so I don't code nasty stuff
use strict;

### SDL Includes
use SDL;
use SDL::App;
use SDL::OpenGL;
use SDL::Mixer;
use SDL::Event;
use Data::Dumper;

### Game libraries
require "data/game.pl";
require "data/glEngine.pl";
require "data/menu.pl";

### Read in the config file
my $rcfile = $ENV{HOME}."/.blockinrc";
my $config = read_config();
my $app;
my $event;

### Main loop
sdl_init();
#main_menu();
game_loop();
sdl_exit();
write_config();

##############################
### Initialize the game engine
##############################
sub sdl_init{
  ### Initialize SDL
  $app = new SDL::App(
    -title 	=> "BlockIn v $version",	### Title of the SDL window
    -width	=> $$config{"width"},		### Width of the screen
    -height	=> $$config{"height"},		### Height of the screen
    -flags 	=> 0,				### FLAGS
    -depth	=> $$config{"depth"},		### Color depth of the screen
    -gl 	=> 1,				### Use OpenGL
    ###-icon 	=> "graphics/icon.png",		### Icon for iconified version
  );
  
  ### Initialize OpenGL
  glViewport(0,0,$$config{"width"},$$config{"height"});	### Set up viewing window
  glMatrixMode(GL_PROJECTION());			### Change to projection mode
  glLoadIdentity();					### Load the default matrix
  glShadeModel(GL_SMOOTH());				### Set our shading model
  glEnable(GL_BLEND);					### Enable blending
  glClearDepth(1.0);					### Set the clear depth
  glDepthFunc(GL_LESS);					### Use the GL_LESS Depth Function
  glEnable(GL_DEPTH_TEST);				### Enable the depth testing

  ### Input init
  $event = new SDL::Event;				### Setup the event queue
  $event->set(SDL_SYSWMEVENT,SDL_IGNORE);		### Ignore WM inputs
}

###############################
### Load the configuration file
###############################
sub read_config{
  ### Check if rcfile exists, and open config file
  if(-e $rcfile){
    my %config;
    ### Open the settings file 
    open RC, $rcfile or die "Can't open $rcfile for reading!\n";
    ### Get the settings
    my @conf;
    while(<RC>){
      @conf = split /=/;
      chomp($conf[1]);
      $config{$conf[0]} = $conf[1];
    }
    close RC;
    return \%config;
  }
  ### Else pass defaults
  else{
    my $config = {
      width 		=> 640,
      height		=> 480,
      depth		=> 16,
      shapeset		=> "basic",
      junk		=> 0,
      field_depth 	=> 8,
      field_length 	=> 8,
      field_width 	=> 8,
      keyrate		=> 75,
      level		=> 1000,
    };
    return $config;
  }
}

############################
### Write configuration file
############################
sub write_config{
  open RC, ">".$rcfile or die "Can't open $rcfile for writing!\n";
  foreach(keys %$config){
    print RC $_."=".$$config{$_}."\n";
  }
  close RC;
}

#######################
### Start the main menu
#######################
sub main_menu{  
  ### Start a new menu
  my $menu = new Menu(title=>"BlockIn v$version",app=>$app,event=>$event,config=>$config);
  ### Start a 1 Player Game
  $menu->add_item(new MenuItem(text=>"Start Game",action=>\&game_menu));
  ### Start a Network Game
  #$menu->add_item(new MenuItem(text=>"Network Game");
  ### High Scores
  #$menu->add_item(new MenuItem(text=>"High Scores");
  ### Options
  #$menu->add_item(new MenuItem(text=>"Options");
  ### Exit
  $menu->add_item(new MenuItem(text=>"Exit",action=>sub { exit(0); }));
  ### Start the menu
  $menu->do_menu();
}

##############################
### Start a 1 Player Game Menu
##############################
sub game_menu{
  my $menu = new Menu(title=>"Single Player Game",app=>$app,event=>$event,config=>$config);
  ### Speed Level X
  
  ### Trash Level	X
  ### Shape Set	X
  ### Start the game
  $menu->add_item(new MenuItem(text=>"Start the Game",action=>\&game_loop));
  ### Main Menu
  $menu->add_item(new MenuItem(text=>"Main Menu"));
}

##############################
### Start a 2 Player Game Menu
##############################
### Server or Client?

##########
### Client
##########
### Server Address: 
### Start Game

##########
### Server
##########
### Speed Level X
### Trash Level	X
### Shape Set	X
### Start Game

#############
### Game Loop
#############
sub game_loop{
  ### Set up the game instance
  my $game = new Game(
    shapeset	=> $$config{shapeset},
    width	=> $$config{field_width},
    length	=> $$config{field_length},
    depth	=> $$config{field_depth},
    #junk	=> 3,
  );
    
  ### Game Loop
  my $done = 0;
  my $keydown = 0;
  my $timer = $app->ticks + 1000;
  while(!$done){ 
    ### Clear the screen
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    ### Reset the view
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
  
    ### Set the perspective
    gluPerspective(100.0, $$config{width}/$$config{height}, 1.0, $$config{field_depth}*2);
    gluLookAt($$config{field_width}/2, $$config{field_length}/2, -$$config{field_depth}/2,
              $$config{field_width}/2, $$config{field_length}/2, $$config{field_depth},
	      0.0, 1.0, 0.0);
  
    ### Draw graphics
    $game->draw();
    $app->sync();
    
    ### Check input
    $event->pump();
    $event->poll();
   
    ### Check for window closes
    exit(0) if($event->type == SDL_QUIT);
  
    ### Check for the key press
    if($event->type == SDL_KEYDOWN && $keydown == 0){
      ### Check for quits
      $done = 1 if($event->key_sym == SDLK_ESCAPE);
      ### Check for piece moves
      $game->{piece}->move(1,0,0,$game->{field})  if ($event->key_sym == SDLK_LEFT);
      $game->{piece}->move(-1,0,0,$game->{field}) if ($event->key_sym == SDLK_RIGHT);
      $game->{piece}->move(0,1,0,$game->{field})  if ($event->key_sym == SDLK_UP);
      $game->{piece}->move(0,-1,0,$game->{field}) if ($event->key_sym == SDLK_DOWN);
     
      ### Check for piece rotations
      $game->{piece}->rotate(\&rotate_x, 1,$game->{field}) if ($event->key_sym == SDLK_q);
      $game->{piece}->rotate(\&rotate_x,-1,$game->{field}) if ($event->key_sym == SDLK_a);
      $game->{piece}->rotate(\&rotate_y, 1,$game->{field}) if ($event->key_sym == SDLK_w);
      $game->{piece}->rotate(\&rotate_y,-1,$game->{field}) if ($event->key_sym == SDLK_s);
      $game->{piece}->rotate(\&rotate_z, 1,$game->{field}) if ($event->key_sym == SDLK_e);
      $game->{piece}->rotate(\&rotate_z,-1,$game->{field}) if ($event->key_sym == SDLK_d);

      ### Check for piece drops
      $timer = $app->ticks() - $$config{level} if ($event->key_sym == SDLK_SPACE);
     
      ### Set the timer for the next keypress
      $keydown = $app->ticks() + $$config{keyrate};
    }
    ### Check for the key release
    elsif($event->type == SDL_KEYUP || $keydown < $app->ticks()){
      $keydown = 0;
    }
    ### Check for timer drops
    if($app->ticks() > $timer){
      if($game->{field}->check_piece_settled($game->{piece})){
	$game->{field}->drop_piece($game->{piece});
        $done = 1 if(!$game->get_random_piece());
      }
      else{ 
        $game->{piece}->{z}++;
      }
      $timer = $app->ticks + $$config{level};
    }
   
    ### Frame Sync
  
    
  }
  ### Display "You Lose" Message and wait for the user to press a key
  print "You lose!\n";

  ### Display scores
  print "Score: \n";

}

### Exit the game engine
sub sdl_exit{

}
