#############################################################################
### BlockIn Game Parent Class
### Copyright 2004 Joseph Jezak
#############################################################################

#############################################################################
### This file is part of BlockIn.
###
### BlockIn is free software; you can redistribute it and/or modify
### it under the terms of the GNU General Public License as published by
### the Free Software Foundation; either version 2 of the License, or
### (at your option) any later version.
###
### BlockIn is distributed in the hope that it will be useful,
### but WITHOUT ANY WARRANTY; without even the implied warranty of
### MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
### GNU General Public License for more details.
###
### You should have received a copy of the GNU General Public License
### along with BlockIn; if not, write to the Free Software
### Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

sub Game::new{
  my $class = shift;
  my %op = @_;
  my $self = {
    shapes	=> 0,			### Possible shapes
    players	=> 0,			### Player data
    field	=> 0,			### Field data
    piece	=> 0, 			### Current piece in play
    level	=> 0,			### Controls the difficulty
  };
  bless $self,$class;
  ### Set up the field
  $self->{field} = new Field(width=>$op{width},length=>$op{length},depth=>$op{depth});
  $self->{field}->junkup($op{junk}) if($op{junk});
  ### Get shape data
  $self->{shapes} = load_shapes($op{shapeset},0);
  $self->get_random_piece();
  return $self;
}

#############################
### Draws the game components
#############################
sub Game::draw{
  my ($self) = @_;
  ### Draw the current piece
  $self->{piece}->draw();
  ### Draw the field
  $self->{field}->draw();
}

##########################
### Get a new piece
##########################
sub Game::get_random_piece{
  my ($self) = @_;
  my $shapes = $self->{shapes};
  ### Remove old piece
  undef($self->{piece});
  ### Create new on
  my $piece = new Piece($$shapes[int(rand $#$shapes+1)]->copy());
  ### Check for end game
  return 0 if($self->{field}->check_piece_settled($piece));
  ### Place the piece on the board
  $piece->{z} = 0;
  $self->{piece} = $piece;
  return 1;
}

##########################
### Player
##########################
sub Player::new{
  my $class = shift;
  my %op = @_;
  my $self = {
    name	=> $op{name},		### Name of the player
    time	=> 0,			### Time played
    score 	=> 0,			### Player's score
    pieces	=> 0,			### Pieces dropped this time
    piecestot	=> 0,			### Total pieces dropped
    levels	=> 0,			### Levels completed
    levelstot	=> 0,			### Total levels completed
  };
  bless $self, $class;
}

###########################
### Game field
###########################
sub Field::new{
  my $class = shift;
  my %op = @_;
  my $self = {
    field	=> 0,
    fieldbg	=> glGenLists(1),	### Background grid Display List
    fieldblocks => glGenLists(1),	### Blocks in the field
    width	=> $op{width},  	### Length of the field
    length	=> $op{length},		### Width of the field
    depth	=> $op{depth},	 	### Height of the field
  };
  ### Create the field
  my @field;
  ### Cycle through the level
  for(my $i=0;$i<$self->{depth};$i++){
    ### And through the columns in every level
    my @length;
    for(my $j=0;$j<$self->{length};$j++){
      ### This is the row bitvector
      $length[$j] = 0;
    }
    $field[$i] = \@length;
  }
  $self->{field} = \@field; 
 
  ### Generate the display list for the field's background
  ### FIXME CHANGE 0x1300 TO GL_COMPILE WHEN I GET A VERSION OF SDL PERL THAT HAS IT
  glNewList($self->{fieldbg},0x1300); 
    draw_background($self->{width},$self->{length},$self->{depth});
  glEndList();
  bless $self, $class;
}

##################################
### Check if the piece has settled
### Returns 1 if yes, 0 if not
##################################
sub Field::check_piece_settled{
  my ($self, $piece) = @_;
  ### Loop through the cubes and check each for a hit
  my $cubes = $piece->{shape}->{cubes};
  foreach(@$cubes){
    ### Check if there is a piece in the spot below or if we've reached the bottom
    return 1 if($piece->{z} + $_->{z} - 1 == $self->{depth} - 2 ||
      $self->check($piece->{x} + $_->{x}, $piece->{y} + $_->{y}, $piece->{z} + $_->{z} + 1));
  }
  return 0;
}

##################################
### Puts the pieces into the field
### Returns # of levels completed
##################################
sub Field::drop_piece{
  my ($self, $piece) = @_;
  ### Loop through the cubes, placing them into the field
  my $cubes = $piece->{shape}->{cubes};
  foreach(@$cubes){
    $self->fill($piece->{x} + $_->{x},$piece->{y} + $_->{y},$piece->{z} + $_->{z});
  }
 
  ### Check for completed levels
  my $filled_levels = 0;
  my $levels = $self->{field};
  for(my $i=0;$i<$self->{depth};$i++){
    my $level = $$levels[$i];
    my $sum = 0;
    foreach(@$level){
      $sum = $sum + $_;
    }
    ### Check if the level is filled
    if($sum == (2**($self->{width}) - 1) * $self->{length}){
      ### Splice out the old level 
      splice @$levels, $i, 1;
      ### Create a new empty level
      my @new_level;
      for(my $j=0;$j<$self->{length};$j++){
      	### This is the row bitvector
      	$new_level[$j] = 0;
      }
      ### Push in the new level
      unshift @$levels,\@new_level; 
      $i--;
      $filled_levels++;
    }
  }
  
  ### Update the field
  $self->update();
  return $filled_levels;
}

##################################
### Fill a spot in the field
### void return
##################################
sub Field::fill{
  my ($self,$x,$y,$z) = @_;
  ### Get the fiels ref
  my $field = $self->{field};
  ### Get the level ref
  my $level = $$field[$z];
  ### Fill the spot
  $$level[$y] = $$level[$y] | (2**$x);
}

##########################################################################
### Fill up the field with junk
### Need to change this so that it checks if there are empty rows above
### the topmost piece.  Then, splice out the empty levels and add in
### junked levels.  This will allow code re-use for multiplayer
##########################################################################
sub Field::junkup{
  my ($self, $junklevel) = @_;

  ### Loop through the field, filling it with junk
  my $levels = $self->{field};
  for(my $i=$self->{depth}-1;$i>($self->{depth} - 1) - $junklevel;$i--){
    my $level = $$levels[$i];
    ### Loop through the columns
    for(my $j=0;$j<$self->{length};$j++){
      $$level[$j] = int(rand 2**($self->{width}));  
    }
    $$levels[$i] = $level;
  }
  ### Draw the field
  $self->update();
}

###################################################
### Check a point and see if it's got a piece in it
###################################################
sub Field::check{
  my ($self, $x, $y, $z) = @_;
  ### Get the fields ref
  my $field = $self->{field};
  ### Get the level ref
  my $level = $$field[$z];
  ### Return the y/n response
  return ($$level[$y] & (2**$x));
}

#####################################
### Check if a point is in bounds
#####################################
sub Field::check_bounds{
  my ($self, $x, $y, $z) = @_;
  ### Check X Axis
  return 0 if($x >= $self->{width} || $x < 0);
  ### Check Y Axis
  return 0 if($y >= $self->{length} || $y < 0);
  ### Check Z Axis
  return 0 if($z >= $self->{depth}); ### Piece can rotate out of the top? || $z < 0);

  ### In bounds
  return 1;
}

###########################
### Draws the field
###########################
sub Field::draw{
  my ($self) = @_;
  ### Draw the background
  glCallList($self->{fieldbg});
  ### Draw the cubes inside the field
  glCallList($self->{fieldblocks});
}

####################
### Update the field
####################
sub Field::update{
  my ($self) = @_;
  glNewList($self->{fieldblocks},0x1300); 
    draw_field($self);  
  glEndList();
}

###########################
### Piece currently in play
###########################
sub Piece::new{
  my ($class,$shape) = @_;
  my $self = {
    shape	=> $shape,		### Shape pointed to
    rot_x	=> 0,			### Rotation on the X plane
    rot_y	=> 0,			### Rotation on the Y plane
    rot_z	=> 0,			### Rotation on the Z plane
    x		=> 0,			### x location
    y		=> 0,			### y location
    z		=> -1,			### z location
  };
  bless $self, $class;
  return $self;
}

########################################
### Draw the piece
########################################
sub Piece::draw{
  my ($self) = @_;
  ### Move the piece
  glPushMatrix();
  glTranslate($self->{x},$self->{y},$self->{z});
  
  ### Rotate piece
  glRotate($self->{rot_x}, 1.0, 0.0, 0.0);
  glRotate($self->{rot_y}, 0.0, 1.0, 0.0);
  glRotate($self->{rot_z}, 0.0, 0.0, 1.0);
 
  ### Draw the piece
  $self->{shape}->draw(1,1);

  glPopMatrix;
}

#############################################
### Rotate the piece in the $rotate direction
#############################################
sub Piece::rotate{
  my ($self, $rotate, $direction,$field) = @_;
  my $cubes = $self->{shape}->{cubes};
  ### Rotate the cubes
  my @rot_cubes;
  my $rot; 
  foreach(@$cubes){
    ### Do the rotation
    $rot = &$rotate($_, $direction);
    ### If there is no piece where we moved to and we're not out of bounds
    if($field->check_bounds($rot->{x} + $self->{x},$rot->{y} + $self->{y},$rot->{z} + $self->{z}) && 
    	!$field->check($rot->{x} + $self->{x},$rot->{y} + $self->{y},$rot->{z} + $self->{z})){ 
      ### Push the rotated cube into the new cube array
      push @rot_cubes, $rot;
    }
  }
  ### Check if we rotated all of the cubes
  $self->{shape}->{cubes} = \@rot_cubes if($#rot_cubes == $#$cubes);
}

##############################################
### Move the piece in the direction specified
##############################################
sub Piece::move{
  my ($self, $x, $y, $z, $field) = @_;
  my $cubes = $self->{shape}->{cubes};
  foreach(@$cubes){
    ### Check for a piece where we want to go and out of bounds
    if(!$field->check_bounds($x + $_->{x} + $self->{x}, $y + $_->{y} + $self->{y}, $z + $_->{z} + $self->{z}) ||
       $field->check($x + $_->{x} + $self->{x}, $y + $_->{y} + $self->{y}, $z + $_->{z} + $self->{z})){
      return;
    }
  }
 
  ### Apply move
  $self->{x} += $x;
  $self->{y} += $y;
  $self->{z} += $z;
}

######################
### Shape object
######################
sub Shape::new{
  my ($class,$name,$cubes) = @_;
  my @cubes;
  my $self = {
    name	=> $name,				### Name of the shape
    cubes	=> $cubes || \@cubes,			### Blocks that make up the shape
  };
  bless $self,$class;
  return $self;
}

############################
### Adds a cube to the shape
############################
sub Shape::add_cube{
  my ($self,$point) = @_;
  my $cubes = $self->{cubes};
  ### Create the vector
  my $vec = new Vector(vector=>$point);
  push @$cubes, $vec;
}

########################
### Copy a shape object
########################
sub Shape::copy{
  my ($self) = @_;
  my $cubes = $self->{cubes};
  my @newcubes = @$cubes;
  return new Shape($self->{name},\@newcubes);
}

#######################
### Draw a shape object
#######################
sub Shape::draw{
  my ($self, $wireframe, $color) = @_;
  my $cubes = $self->{cubes};
  foreach(@$cubes){
    draw_cube($_->{x},$_->{y},$_->{z},$wireframe,$color);
  }
}

#################
### 3 Item Vector
#################
sub Vector::new{
  my $class = shift;
  my %op = @_;
  ### Allow for packed vectors
  if(exists($op{"vector"})){
    my @vector = split /,/,$op{"vector"};
    $op{"x"} = $vector[0];
    $op{"y"} = $vector[1];
    $op{"z"} = $vector[2];
  }
  ### Set up the vector
  my $self = {
    x		=> $op{"x"} || 0,
    y		=> $op{"y"} || 0,
    z		=> $op{"z"} || 0,
  };
  bless $self, $class;
  return $self;
}

################
#### Vector Copy
################
sub Vector::copy{
  my ($self) = @_;
  return new Vector(x=>$self->{x},y=>$self->{y},z=>$self->{z});
}

1;
